/**
 * Subor sprintf.c
 *
 * Zjednoduseny sprintf (kvoli redukcii velkosti kodu)
 * 
 * This file can be compiled into more than one flavour.  The default
 * is to offer the usual modifiers and integer formatting support
 * (level 2).  Level 1 maintains a minimal version that just offers
 * integer formatting, but no modifier support whatsoever.  Level 3 is
 * intented for floating point support.
 * 
 * Poznamky:
 * Kod prevzaty z avr-libc. Zalozene na vfprintf.c, verzia 1.13
 * 
 * Velkost kodu:
 * - MIN 916 bajtov
 * - STD 1492 bajtov
 * - FLT 6752 bajtov
 * 
 */
 
/* Copyright (c) 2002, Alexander Popov (sasho@vip.bg)
   Copyright (c) 2002,2004,2005 Joerg Wunsch
   Copyright (c) 2005, Helmut Wallner
   Copyright (c) 2008, Michael Wolf
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions are met:

   * Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
   * Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the
     distribution.
   * Neither the name of the copyright holders nor the names of
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
  POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdint.h>
#include <limits.h>
#include <stdarg.h>
#include <stdlib.h>
#include "retazec.h"
#include "sprintf.h"

// Hodnoty pre PRINTF_LEVEL
#define PRINTF_MIN 1
#define PRINTF_STD 2
#define PRINTF_FLT 3

#ifndef PRINTF_LEVEL
#  define PRINTF_LEVEL PRINTF_STD
#endif

#if PRINTF_LEVEL == PRINTF_MIN || PRINTF_LEVEL == PRINTF_STD \
|| PRINTF_LEVEL == PRINTF_FLT
/* OK */
#else
#  error "Neznama uroven."
#endif

#if PRINTF_LEVEL >= PRINTF_FLT
#include <math.h>
#endif

// *****************************************************
// Printf uroven MIN
// *****************************************************
#if PRINTF_LEVEL >= PRINTF_MIN
#define FLHASPERCENT    0x01    //!< first % found
#define FLNEGATIVE  0x02        //!< arg is negative int
#define FLLONG      0x04        //!< arg is long int
#endif

// *****************************************************
// Printf uroven STD
// *****************************************************
#if PRINTF_LEVEL >= PRINTF_STD
#define FLPREC      0x08    //!< .prec given
#define FLSIGNPLUS  0x10    //!< always print '+' sign
#define FLSIGN      0x20    //!< always print sign
#define FLALT       0x40    //!< # alternate specification given
#define FLLPAD      0x80    //!< left-align result
#define FLZFILL     0x100   //!< zero-fill up to width
#endif

// *****************************************************
// Printf uroven FLT
// *****************************************************
#if PRINTF_LEVEL >= PRINTF_FLT
#define FLFLOAT     0x200   //!< floating-point conversion
#define FLFCVT      0x400   //!< do style `f' if set, style `e' if clear
#define FLGCVT      0x800   //!< %g format, decide about %e or %f
#define FLTBUFLEN 40    //!< With 4-byte floats, we can get up to 39 digits.

static int8_t   flcvt(char *buf, double arg, int8_t ndigits, uint16_t *flags);

#endif /* PRINTF_LEVEL >= PRINTF_FLT */

#define BUFLEN 12   //!< Integer conversion buffer length.

// *****************************************************
// Redukovana funkcia printf
// *****************************************************
void rsprintf(char *stream, const char *fmt, ...) {
    union {
        char c;
        uint8_t u8;
        int8_t i8;
        long l;
        unsigned long ul;
        char *pc;
#if PRINTF_LEVEL >= PRINTF_FLT
        double d;
        div_t dv;
#endif
    } a;
    
    va_list ap;
    char    c=0;  /* znak z formatovacieho retazca */
    uint8_t base=0;
    char    b[BUFLEN]; /* Obsahuje znakovu reprezentaciu cisla a prefix odzadu dopredu */
    char    *pb=NULL;  /* smernik na polozku v b */
#if PRINTF_LEVEL > PRINTF_MIN
    int8_t  width=0, prec=0;
#endif
#if PRINTF_LEVEL >= PRINTF_FLT
    int8_t  decpt;
    char    fb[FLTBUFLEN];  /* buffer pre float */
#endif

#if PRINTF_LEVEL < PRINTF_STD
    uint8_t flags;
#else
    uint16_t flags;
#endif

    flags = 0;
    
    va_start (ap, fmt);
    
    while ((c = *fmt)) {

        fmt++;

        if (flags & FLHASPERCENT) {
            if (c >= '0' && c <= '9') {
#if PRINTF_LEVEL > PRINTF_MIN
                if (c == '0' && prec == 0 &&
                    (flags & FLPREC) == 0) {
                    flags |= FLZFILL;
                } else {
                    /* digit */
                    prec *= 10;
                    prec += c - '0';
                    if (prec < 0)
                        prec = SCHAR_MAX;
                    if (!(flags & FLPREC))
                        width = prec;
                }
#endif
#if PRINTF_LEVEL >= PRINTF_FLT
            } else if ((c >= 'e' && c <= 'g') ||
                   (c >= 'E' && c <= 'G')) {
                a.d = va_arg(ap, double);
                if (!(flags & FLPREC))
                    prec = 6;
                /* FLPREC je pouzity len pre cele cisla */
                flags &= ~(FLALT | FLPREC);
                if (isnan(a.d)) {
                    b[0] = b[2] = 'N';
                    b[1] = 'A';
                    pb = b + 3;
                  dofltexcpt:
                    if (c >= 'e' && c <= 'g') {
                        b[0] += 'a' - 'A';
                        b[1] += 'a' - 'A';
                        b[2] += 'a' - 'A';
                    }
                    goto nextitem;
                } else if (isinf(a.d)) {
                    b[2] = 'I';
                    b[1] = 'N';
                    b[0] = 'F';
                    pb = b + 3;
                    if (a.d < 0.0)
                        *pb++ = '-';
                    goto dofltexcpt;
                }
                flags |= FLFLOAT;
                switch (c) {
                case 'e':
                case 'E':
                    decpt = flcvt(fb, a.d, prec, &flags);
                  dostyle_e:
                    width -= 4; /* E+00 */
                    if (prec > 0)
                        width--; /* decimal sign */
                    a.i8 = strlen(pb = fb);
                    goto dofloat;

                case 'f':
                case 'F':
                    flags |= FLFCVT;
                    decpt = flcvt(fb, a.d, prec, &flags);
                  dostyle_f:
                    if (prec > 0)
                        width--; /* decimal sign */
                    pb = fb;
                    if (decpt > 0)
                        a.i8 = strlen(pb);
                    else
                        a.i8 = prec + 1;
                    goto dofloat;
                case 'g':
                case 'G':
                    flags |= FLGCVT;
                    decpt = flcvt(fb, a.d,
                              prec == 0? 1: prec,
                              &flags);
                    /*
                     * Odstranenie uvodnych nul z vysledku
                     */
                    if (flags & FLFCVT) {
                        for (a.i8 = decpt + prec - 1;
                             a.i8 >= 0 && a.i8 >= decpt;
                             a.i8--)
                            if (fb[a.i8] != '0')
                                break;
                        if (a.i8 < 0 || a.i8 < decpt)
                            /*
                             * Ziadne cisla nezostali, nezobraz ani '.'
                             */
                            prec = 0;
                        else
                            prec = a.i8 - decpt + 1;
                    } else {
                        /* `e' styl */
                        for (a.i8 = prec;
                             a.i8 > 0;
                             a.i8--)
                            if (fb[a.i8] != '0')
                                break;
                        if (a.i8 == 0)
                            /*
                             * Ziadne cisla nezostali, nezobraz ani '.'
                             */
                            prec = 0;
                        else
                            prec = a.i8;
                    }
                    fb[a.i8 + 1] = '\0';

                    if (flags & FLFCVT)
                        goto dostyle_f;
                    c -= 'G' - 'E';
                    goto dostyle_e;
                }
#endif /* PRINTF_LEVEL >= PRINTF_FLT */
            } else {
#if PRINTF_LEVEL > PRINTF_MIN
                if ((flags & FLPREC) && prec == 0)
                    prec = 1;
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                switch (c) {
                case '+':
#if PRINTF_LEVEL > PRINTF_MIN
                    flags |= FLSIGNPLUS;
                    /* FALLTHROUGH */
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                case ' ':
#if PRINTF_LEVEL > PRINTF_MIN
                    flags |= FLSIGN;
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                    break;
                case '-':
#if PRINTF_LEVEL > PRINTF_MIN
                    flags |= FLLPAD;
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                    break;
                case '#':
#if PRINTF_LEVEL > PRINTF_MIN
                    flags |= FLALT;
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                    break;
                case '.':
#if PRINTF_LEVEL > PRINTF_MIN
                    flags |= FLPREC;
                    prec = 0;
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                    break;
                case 'l':
                    flags |= FLLONG;
                    break;
#if PRINTF_LEVEL < PRINTF_FLT
                case 'e':
                case 'f':
                case 'g':
                    a.c = '?';
#endif /* PRINTF_LEVEL < PRINTF_FLT */
#if PRINTF_LEVEL > PRINTF_MIN
                    width--;
#endif /* PRINTF_LEVEL > PRINTF_MIN */
                    c = 'c';
                    goto nextitem;
                case 'c':
                    /* znak je konvertovany na int pomocou va_arg */
                    a.c = (char)va_arg(ap, int);
#if PRINTF_LEVEL > PRINTF_MIN
                    width--;
#endif
                    goto nextitem;
                case 's':
                    a.pc = va_arg(ap, char *);
#if PRINTF_LEVEL > PRINTF_MIN
                    if (flags & FLPREC)
                      base = strlen(a.pc) < prec ? strlen(a.pc) : prec;
                    else
                        base = strlen(a.pc);
                    width -= base;
#else
                    base = strlen(a.pc);
#endif
                    goto nextitem;
                case 'd':
                case 'i':
                    a.l = flags & FLLONG ?
                        va_arg(ap, long):
                        va_arg(ap, int);
                    if (a.l < 0) {
                        flags |= FLNEGATIVE;
                        a.l = -a.l;
                    }
#if PRINTF_LEVEL > PRINTF_MIN
                    flags &= ~FLALT;
#endif
                    goto processnum;
                case 'o': /* cislo v osmickovej sustava */
                    base = 8;
                    goto getulong;
                case 'p':
#if PRINTF_LEVEL > PRINTF_MIN
                    flags |= FLALT;
#endif
                    c = 'x';
                case 'x':
                case 'X':
                    base = 16;
                case 'u':
                  getulong:
                    a.ul = flags & FLLONG?
                        va_arg(ap, unsigned long):
                        va_arg(ap, unsigned int);
#if PRINTF_LEVEL > PRINTF_MIN
                    flags &= ~(FLSIGNPLUS | FLSIGN);
#endif
                  processnum:
                    pb = b;
                    do {
                        *pb = a.ul % base;
                        *pb = *pb > 9?
                            *pb + c - 'X' + 'A' - 10:
                            *pb + '0';
                        *pb++;
                        a.ul /= base;
                    } while (a.ul);
#if PRINTF_LEVEL > PRINTF_MIN
                    /* dlzka konvertovaneho retazca */
                    a.i8 = pb - b;
#endif
#if PRINTF_LEVEL >= PRINTF_FLT
                  dofloat:
#endif
#if PRINTF_LEVEL > PRINTF_MIN
                    /* vypocet dlzky pre doplnenie 0 */
                    if (flags & (FLNEGATIVE | FLSIGN |
                             FLSIGNPLUS))
                        width--;
                    if (flags & FLALT)
                        width -= base == 16? 2: 1;
                    if (flags & FLPREC) {
                        if (prec <= (uint8_t)a.i8)
                            /*
                             * Ziadny priestor pre doplnujuce 0
                             */
                            flags &= ~FLPREC;
                        else
                            prec -= (uint8_t)a.i8;
                    }
                    if (flags & FLPREC)
                        width -= prec;
                    else
                        width -= a.i8;

                    /* doplnenie zlava ak je potrebne */
                    if (!(flags & (FLLPAD | FLZFILL)))
                        while (width-- > 0)
                            (*stream++)=' ';
#endif
                    if (flags & FLNEGATIVE)
                        (*stream++)='-';
#if PRINTF_LEVEL > PRINTF_MIN
                    else if (flags & FLSIGNPLUS)
                        (*stream++)='+';
                    else if (flags & FLSIGN)
                        (*stream++)=' ';
                    if (flags & FLALT) {
                        (*stream++)='0';
                        if (base == 16)
                            (*stream++)=c;
                    }
                    if ((flags & FLZFILL))
                        while (width-- > 0)
                            (*stream++)='0';
                    if (flags & FLPREC)
                        while (prec-- > 0)
                            (*stream++)='0';
#endif
                  nextitem:
#if PRINTF_LEVEL > PRINTF_MIN
                    /*
                     * Toto je len pre %s and %c
                     */
                    if (!(flags & FLLPAD))
                        while (width-- > 0)
                            (*stream++)=' ';
#endif
                    if (c == 'c')
                        (*stream++)=a.c;
                    else if (c == 's')
                        while (base--)
                            (*stream++)=*a.pc++;
#if PRINTF_LEVEL >= PRINTF_FLT
                    else if (flags & FLFLOAT) {
                        if (flags & FLFCVT) {
                            if (decpt > 0) {
                                while (decpt--)
                                    (*stream++)=*pb++;
                                decpt = 0;
                            } else
                                (*stream++)='0';
                        } else
                            /* 'e' format */
                            (*stream++)=*pb++;
                        if (prec > 0)
                            (*stream++)='.';
                        if (flags & FLFCVT)
                            while (decpt++ < 0 &&
                                   prec-- > 0)
                                    (*stream++)='0';
                        while ((a.c = *pb++) &&
                               prec-- > 0)
                            (*stream++)=a.c;
                        if (!(flags & FLFCVT)) {
                            /* 'e' format */
                            (*stream++)=c;
                            decpt--;
                            if (decpt < 0) {
                                (*stream++)='-';
                                decpt = -decpt;
                            } else
                                (*stream++)='+';
                            a.dv = div(decpt, 10);
                            (*stream++)=a.dv.quot + '0';
                            (*stream++)=a.dv.rem + '0';
                        }
                    }
#endif /* PRINTF_LEVEL >= PRINTF_FLT */
                    else
                        while (pb != b)
                            (*stream++)=*--pb;
#if PRINTF_LEVEL > PRINTF_MIN
                    if (flags & FLLPAD)
                        while (width-- > 0)
                            (*stream++)=' ';
#endif
                    goto clearflags;

                default:
                    (*stream++)=c;
                  clearflags:
                    flags = 0;
                    break;
                }
            }
        } else
            if (c == '%') {
                flags = FLHASPERCENT;
#if PRINTF_LEVEL > PRINTF_MIN
                prec = width = 0;
#endif
                base = 10;
            } else
                (*stream++)=c;
    }

    va_end(ap);
   (*stream)=0;   // ukoncujuci znak retazca
}

#if PRINTF_LEVEL >= PRINTF_FLT

/*
 * Konverzia cisla v plavajucej radovej ciarke na retazec
 *
 * Featured after the Unix fcvt()/ecvt() functions.  The argument is
 * converted into a digit string.
 *
 * For style `f' conversion (*flags & FLFCVT is set), the position of
 * the decimal point is returned, where a negative position means the
 * decimal point is outside our buffer to the left.  The buffer will
 * then start with the first non-zero digit.  For numbers greater then
 * 1, trailing zeros will be added into the buffer though.  prec
 * determines the number of digits after the decimal point.  (Note
 * that the decimal point itself will not be added into the buffer.)
 *
 * For style `e' conversion (*flags & FLFCVT is clear), decpt is
 * effectively the exponent of our conversion minus 1, and the
 * conversion is always done so that the first digit in the conversion
 * buffer is considered to be the digit before the decimal point, with
 * prec digits following.
 *
 * For style `g' conversion (*flags & FLGCVT is set), after
 * determining the exponent of the conversion, it will be decided
 * whether an `e' style or an `f' style conversion needs to be done.
 * In the former case, *flags will have FLFCVT set upon return.
 *
 * For negative numbers, *flags will get FLNEGATIVE set upon return,
 * while the conversion itself will be handled as if a positive number
 * was given.
 */

#include <math.h>

static int8_t
flcvt(char *buf, double arg, int8_t prec, uint16_t *flags)
{
    int8_t decpt;
    double dnum, dnum2;
    char *p, *p1, *pend;

    /*
     * Obmedzenie presnosti
     */
    if (prec < 0)
        prec = 0;
    if (prec > FLTBUFLEN - 3)
        prec = FLTBUFLEN - 3;
    decpt = 0;
    p = buf;
    pend = buf + FLTBUFLEN;
    /*
     * Najprv konverzia na kladne cislo a rozdelenie na celu a desatinnu cast
     */
    if (arg < 0) {
        *flags |= FLNEGATIVE;
        arg = -arg;
    }
    arg = modf(arg, &dnum);
    /*
     * Cela cast. Cisla su zapisane odzadu dopredu.
     */
    if (dnum != 0) {
        for (p1 = pend; dnum != 0 && p1 > p; decpt++) {
            dnum2 = modf(dnum * 0.1, &dnum);
            *--p1 = (int)(dnum2 * 10.1) + '0';
        }
        while (p1 < pend)
            *p++ = *p1++;
    }
    /*
     * Ak cislo bolo mensie nez 0, iba sa posunie do rozsahu [0.1 .. 1.0), a poznaci sa pocet cisel na posunutie
     */
    else if (arg > 0) {
        while ((dnum = arg * 10) < 1) {
            arg = dnum;
            decpt--;
        }
    }
    if (*flags & FLGCVT) {
        if (decpt >= -3 && decpt < prec + 1)
            *flags |= FLFCVT;
    }
    /*
     * `e' styl
     */
    if (!(*flags & FLFCVT))
        prec++;
    p1 = buf + prec;
    if (*flags & FLFCVT)
        p1 += decpt;
    /*
     * `f' styl
     */
    if (p1 < buf) {
        buf[0] = '\0';
        return decpt;
    }
    /*
     * Konverzia prvych cisel desatinnej casti podla pozadovanej presnosti
     */
    while (p <= p1 && p < pend) {
        arg *= 10;
        arg = modf(arg, &dnum);
        *p++ = (int)dnum + '0';
    }
    /*
     * Ak sme na konci buffra, ukonci sa retazec a je hotovo.
     */
    if (p1 >= pend) {
        buf[FLTBUFLEN-1] = '\0';
        return decpt;
    }
    /*
     * Zaokruhlenie vysledku
     */
    p = p1;
    *p1 += 5;
    while (*p1 > '9') {
        *p1 = '0';
        if (p1 > buf)
            ++*--p1;
        else {
            *p1 = '1';
            decpt++;
            if (*flags & FLFCVT) {
                if (p > buf)
                    *p = '0';
                p++;
            }
        }
    }
    *p = '\0';
    return decpt;
}

#endif /* PRINTF_LEVEL >= PRINTF_FLT */
